package skyjed.gui;

import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.*;
import java.io.*;
import java.util.List;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.text.JTextComponent;

import skyjed.buffer.BooleanBuffer;
import skyjed.buffer.BooleanBufferArray;
import skyjed.gui.action.ILoadStoreValues;
import skyjed.gui.action.UpdateSelectedItemsInJListAction;
import skyjed.gui.util.*;
import skyjed.resources.*;
import skyjed.save.SkyPkmn;
import skyjed.save.util.LengthMismatchException;
import skyjed.save.util.SkyBufferUtils;
import skyjed.save.util.SkyClipboardInfo;
import skyjed.util.*;

public class ManagePokemonWindow implements ILoadStoreValues<SkyPkmn> {

    private static final int PKMN_RAW_FILE_DATA_LENGTH = 46;
    private static final int PKMN_SPLIT_FILE_DATA_OFFSET = 6;
    private static final int PKMN_SPLIT_FILE_DATA_LENGTH = 362;

    private static final int PKMN_NAME_LENGTH = 10;
    private static final int PKMN_METFL_MAX = 99;
    private static final int PKMN_LVL_MAX = 100;
    private static final int PKMN_EXP_MAX = 9_999_999;
    private static final int PKMN_HP_MAX = 999;
    private static final int PKMN_IQ_MAX = 999;
    private static final int PKMN_STAT_MAX = 255;

    private static final int PKMN_NUM_ATTACKS = 4;

    private JFrame frame;

    private JList<SkyPkmn> pkmnList;

    private JCheckBox chkBoxIsValid;
    private JCheckBox chkBoxIsFemale;
    private JTextComponent txtName;
    private JComboBox<PropertyItem> cmbSpecies;
    private JComboBox<PropertyItem> cmbMetAt;
    private JNumberSpinner spnMetFloor;
    private JNumberSpinner spnLevel;
    private JNumberSpinner spnExp;
    private JNumberSpinner spnHP;
    private JNumberSpinner spnIQ;
    private JNumberSpinner spnAtk;
    private JNumberSpinner spnSpAtk;
    private JNumberSpinner spnDef;
    private JNumberSpinner spnSpDef;
    private JNumberSpinner spnEvo1Lvl;
    private JNumberSpinner spnEvo2Lvl;
    private JComboBox<PropertyItem> cmbTactic;

    private AttackEditPanel[] attackEditPanels;

    private JBooleanLabel unk1;
    private JBooleanLabel[] iqMapList = new JBooleanLabel[SkyPkmn.LEN_IQ_MAP];

    private Action updatePokemonAction;

    private ActionListener actionListener = new ActionListener() {
        public void actionPerformed(ActionEvent ev) {
            switch(ev.getActionCommand()) {
                case "Import Pokemon":
                {
                    JFileChooser jfc = new JFileChooser(new File("."));
                    int returnval = jfc.showOpenDialog(frame);
                    if(returnval == JFileChooser.APPROVE_OPTION) {
                        importFile(jfc.getSelectedFile());
                    }
                    break;
                }
                case "Export Pokemon":
                {
                    JFileChooser jfc = new JFileChooser(new File("."));
                    int returnval = jfc.showSaveDialog(frame);
                    if(returnval == JFileChooser.APPROVE_OPTION) {
                        exportFile(jfc.getSelectedFile());
                    }
                    break;
                }
            }
        }
    };

    private ListSelectionListener listSelectionListener = new ListSelectionListener(){
        public void valueChanged(ListSelectionEvent ev) {
            if(ev.getValueIsAdjusting()) return;
            loadValues(pkmnList.getSelectedValue());
        }
    };

    private MouseListener listMouseListener = new PopupMouseAdapter() {
        private SkyClipboardInfo clipInfo;

        private ActionListener popupMenuActionListener = new ActionListener() {
            public void actionPerformed(ActionEvent ev) {
                try {
                    List<SkyPkmn> pkmns = pkmnList.getSelectedValuesList();
                    // Nothing selected - do nothing
                    if(pkmns.size() == 0) return;
                    if(pkmns.size() == 1) {
                        // Single selection
                        SkyPkmn pkmn = pkmns.get(0);
                        switch(ev.getActionCommand()) {
                            case "Popup->Clear":
                                SkyBufferUtils.clear(pkmn);
                                break;
                            case "Popup->Cut":
                                SkyBufferUtils.cut(pkmn);
                                break;
                            case "Popup->Copy":
                                SkyBufferUtils.copy(pkmn);
                                break;
                            case "Popup->Paste":
                                if(clipInfo.type == SkyClipboardInfo.TYPE_SINGLE) {
                                    SkyBufferUtils.paste(pkmn);
                                } else { // this is hacky
                                    int idxHead = pkmnList.getSelectedIndex();
                                    int idxTail = idxHead + clipInfo.arrayLength;
                                    if(idxTail > pkmnList.getModel().getSize()) {
                                        idxTail = pkmnList.getModel().getSize();
                                    }
                                    pkmnList.setSelectionInterval(idxHead, idxTail-1);
                                    try {
                                        SkyBufferUtils.pasteList(pkmnList.getSelectedValuesList());
                                    } finally {
                                        pkmnList.setSelectedIndex(idxHead);
                                    }
                                }
                                break;
                        }
                    } else {
                        // Multiple selection
                        switch(ev.getActionCommand()) {
                            case "Popup->Clear":
                                SkyBufferUtils.clearList(pkmns);
                                break;
                            case "Popup->Cut":
                                SkyBufferUtils.cutList(pkmns);
                                break;
                            case "Popup->Copy":
                                SkyBufferUtils.copyList(pkmns);
                                break;
                            case "Popup->Paste":
                                if(clipInfo.type == SkyClipboardInfo.TYPE_ARRAY) {
                                    SkyBufferUtils.pasteList(pkmns);
                                } else {
                                    for(SkyPkmn pkmn : pkmns) {
                                        SkyBufferUtils.paste(pkmn);
                                    }
                                }
                                break;
                        }
                    }
                    // refresh the list
                    pkmnList.repaint();
                } catch (LengthMismatchException e) {
                    ExceptionHandler.handleException(e, frame);
                }
            }
        };

        protected void popUp(MouseEvent e) {
            int index = pkmnList.locationToIndex(e.getPoint());
            if(pkmnList.isSelectedIndex(index) == false) {
                pkmnList.setSelectedIndex(index);
            }

            JPopupMenu mnPopup = new JPopupMenu();

            JMenuItem mntmClear = new JMenuItem("Clear");
            mntmClear.setActionCommand("Popup->Clear");
            mntmClear.addActionListener(this.popupMenuActionListener);
            mnPopup.add(mntmClear);

            JMenuItem mntmCut = new JMenuItem("Cut");
            mntmCut.setActionCommand("Popup->Cut");
            mntmCut.addActionListener(this.popupMenuActionListener);
            mnPopup.add(mntmCut);

            JMenuItem mntmCopy = new JMenuItem("Copy");
            mntmCopy.setActionCommand("Popup->Copy");
            mntmCopy.addActionListener(this.popupMenuActionListener);
            mnPopup.add(mntmCopy);

            JMenuItem mntmPaste = new JMenuItem("Paste");
            clipInfo = SkyBufferUtils.getInfoForMime(SkyPkmn.MIME_TYPE);
            mntmPaste.setEnabled(clipInfo.type != SkyClipboardInfo.TYPE_NONE);
            mntmPaste.setActionCommand("Popup->Paste");
            mntmPaste.addActionListener(this.popupMenuActionListener);
            mnPopup.add(mntmPaste);

            mnPopup.show(e.getComponent(), e.getX(), e.getY());
        }
    };

    public void loadValues(SkyPkmn pkmn) {
        if(pkmn == null) {
            pkmn = new SkyPkmn();
        }
        chkBoxIsValid.setSelected(pkmn.isvalid);
        chkBoxIsFemale.setSelected(pkmn.isfemale);
        txtName.setText(pkmn.name.getString());
        cmbSpecies.setSelectedIndex(pkmn.no);
        cmbMetAt.setSelectedIndex(pkmn.metat);
        spnMetFloor.setValue(pkmn.metfl);
        spnLevel.setValue(pkmn.lvl);
        spnExp.setValue(pkmn.exp);
        spnHP.setValue(pkmn.hp);
        spnIQ.setValue(pkmn.iq);
        spnAtk.setValue(pkmn.stats[0]);
        spnSpAtk.setValue(pkmn.stats[1]);
        spnDef.setValue(pkmn.stats[2]);
        spnSpDef.setValue(pkmn.stats[3]);
        spnEvo1Lvl.setValue(pkmn.evo1lvl);
        spnEvo2Lvl.setValue(pkmn.evo2lvl);
        cmbTactic.setSelectedIndex(pkmn.tactic);
        for(int i=0; i<PKMN_NUM_ATTACKS; i++) {
            attackEditPanels[i].setAttack(pkmn.attacks[i]);
        }
        unk1.setBool(pkmn.unk1);
        for(int i=0; i<iqMapList.length; i++) {
            iqMapList[i].setBool(pkmn.iqmap[i]);
        }
    }

    public void storeValues(SkyPkmn pkmn) {
        // Nothing selected - do nothing
        if(pkmn == null) return;
        pkmn.isvalid = chkBoxIsValid.isSelected();
        pkmn.isfemale = chkBoxIsFemale.isSelected();
        pkmn.name.setString(txtName.getText());
        pkmn.no = cmbSpecies.getSelectedIndex();
        pkmn.metat = cmbMetAt.getSelectedIndex();
        pkmn.metfl = spnMetFloor.getValueInt();
        pkmn.lvl = spnLevel.getValueInt();
        pkmn.exp = spnExp.getValueInt();
        pkmn.hp = spnHP.getValueInt();
        pkmn.iq = spnIQ.getValueInt();
        pkmn.stats[0] = spnAtk.getValueInt();
        pkmn.stats[1] = spnSpAtk.getValueInt();
        pkmn.stats[2] = spnDef.getValueInt();
        pkmn.stats[3] = spnSpDef.getValueInt();
        pkmn.evo1lvl = spnEvo1Lvl.getValueInt();
        pkmn.evo2lvl = spnEvo2Lvl.getValueInt();
        pkmn.tactic = cmbTactic.getSelectedIndex();
        for(int i=0; i<PKMN_NUM_ATTACKS; i++) {
            pkmn.attacks[i] = attackEditPanels[i].getAttack();
        }
        pkmn.unk1 = unk1.getBool();
        for(int i=0; i<iqMapList.length; i++) {
            pkmn.iqmap[i] = iqMapList[i].getBool();
        }
    }

    protected void importFile(File selectedFile) {
        // TODO import multiple?
        try (FileInputStream fin = new FileInputStream(selectedFile)) {
            byte[] rawbuf = new byte[PKMN_RAW_FILE_DATA_LENGTH];
            if(fin.read(rawbuf) < PKMN_RAW_FILE_DATA_LENGTH) {
                throw new IOException("File is too short for a pkmn data");
            }
            BooleanBuffer buf = new BooleanBufferArray(BitConverterLE.splitBits(rawbuf))
            .seek(PKMN_SPLIT_FILE_DATA_OFFSET)
            .view(PKMN_SPLIT_FILE_DATA_LENGTH);
            loadValues(new SkyPkmn(buf));
        } catch (Exception e) {
            ExceptionHandler.handleException(e, frame);
        }
    }

    protected void exportFile(File selectedFile) {
        // TODO export multiple?
        boolean[] splitbuf = new boolean[PKMN_SPLIT_FILE_DATA_OFFSET+PKMN_SPLIT_FILE_DATA_LENGTH];
        SkyPkmn pkmn = new SkyPkmn();
        storeValues(pkmn);
        BooleanBuffer buf = new BooleanBufferArray(splitbuf);
        pkmn.store(buf.seek(PKMN_SPLIT_FILE_DATA_OFFSET).view(PKMN_SPLIT_FILE_DATA_LENGTH));
        byte[] rawbuf = BitConverterLE.packBits(splitbuf);
        try (FileOutputStream fout = new FileOutputStream(selectedFile)) {
            fout.write(rawbuf);
        } catch (Exception e) {
            ExceptionHandler.handleException(e, frame);
        }
    }

    /**
     * Create the window.
     */
    public ManagePokemonWindow(SkyPkmn[] pkmnArray) {
        pkmnList = new JList<SkyPkmn>();
        pkmnList.setModel(new ArrayListModel<SkyPkmn>(pkmnArray));
        updatePokemonAction = new UpdateSelectedItemsInJListAction<>("Update Pokemon", pkmnList, this);
        initialize();
        frame.setVisible(true);
    }

    /**
     * Initialize the contents of the frame.
     */
    private void initialize() {
        frame = new JFrame();
        frame.setTitle("Manage Pokemon Roster");
        frame.setResizable(false);
        frame.setSize(530, 510);
        frame.setLocationByPlatform(true);
        frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

        JPanel contentPane = new GBPanel();
        frame.setContentPane(contentPane);
        {
            GBLayout gbl = new GBLayout();
            gbl.columnWidths(0, 70, 100, 80, 100);
            gbl.rowHeights(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            gbl.columnWeights(1.0, 0.0, 0.0, 0.0, 0.0);
            gbl.rowWeights(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
            contentPane.setLayout(gbl);
        }
        {
            pkmnList.addListSelectionListener(listSelectionListener);
            pkmnList.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
            pkmnList.addMouseListener(listMouseListener);
            contentPane.add(new JScrollPane(pkmnList), new GBConstraints(0, 0, 1, GBConstraints.REMAINDER).fillBoth());
        }
        {
            JPanel chkBoxPanel = new JPanel();
            FlowLayout fl_chkBoxPanel = (FlowLayout) chkBoxPanel.getLayout();
            fl_chkBoxPanel.setVgap(0);
            fl_chkBoxPanel.setHgap(0);
            contentPane.add(chkBoxPanel, new GBConstraints(1, 0, 4, 1).fillBoth());
            chkBoxIsValid = new JCheckBox("Is Valid");
            chkBoxPanel.add(chkBoxIsValid);
            chkBoxPanel.add(Box.createHorizontalStrut(20));
            chkBoxIsFemale = new JCheckBox("Is Female");
            chkBoxPanel.add(chkBoxIsFemale);
        }
        contentPane.add(new JLabel("Species:"), new GBConstraints(1, 1).anchorEast());
        {
            cmbSpecies = new JComboBox<PropertyItem>();
            cmbSpecies.setModel(new ArrayListModel<PropertyItem>(SkyPokemonRes.getInstance().getPropertyList()));
            contentPane.add(cmbSpecies, new GBConstraints(2, 1, 3, 1).fillHorizontal());
        }
        contentPane.add(new JLabel("Met At:"), new GBConstraints(1, 2).anchorEast());
        {
            cmbMetAt = new JComboBox<PropertyItem>();
            cmbMetAt.setModel(new ArrayListModel<PropertyItem>(SkyLocationRes.getInstance().getPropertyList()));
            contentPane.add(cmbMetAt, new GBConstraints(2, 2, 3, 1).fillHorizontal());
        }
        contentPane.add(new JLabel("Name:"), new GBConstraints(1, 3).anchorEast());
        {
            txtName = new JSkyTextField(PKMN_NAME_LENGTH);
            contentPane.add(txtName, new GBConstraints(2, 3).fillHorizontal());
        }
        contentPane.add(new JLabel("Met At Floor:"), new GBConstraints(3, 3).anchorEast());
        spnMetFloor = new JNumberSpinner(PKMN_METFL_MAX);
        contentPane.add(spnMetFloor, new GBConstraints(4, 3).fillHorizontal());
        {
            JPanel statsPanel = new GBPanel();
            contentPane.add(statsPanel, new GBConstraints(1, 4, 4, 1).fillBoth());
            {
                GBLayout gbl = new GBLayout();
                gbl.columnWidths(70, 100, 80, 100);
                gbl.rowHeights(0, 0, 0, 0, 0, 0);
                gbl.columnWeights(0.0, 0.0, 0.0, 0.0);
                gbl.rowWeights(1.0, 0.0, 0.0, 0.0, 0.0, 0.0);
                statsPanel.setLayout(gbl);
            }
            statsPanel.add(new JLabel("Level:"), new GBConstraints(0, 1).anchorEast());
            spnLevel = new JNumberSpinner(PKMN_LVL_MAX);
            statsPanel.add(spnLevel, new GBConstraints(1, 1).fillHorizontal());

            statsPanel.add(new JLabel("Exp:"), new GBConstraints(2, 1).anchorEast());
            spnExp = new JNumberSpinner(PKMN_EXP_MAX);
            statsPanel.add(spnExp, new GBConstraints(3, 1).fillHorizontal());

            statsPanel.add(new JLabel("HP:"), new GBConstraints(0, 2).anchorEast());
            spnHP = new JNumberSpinner(PKMN_HP_MAX);
            statsPanel.add(spnHP, new GBConstraints(1, 2).fillHorizontal());

            statsPanel.add(new JLabel("IQ:"), new GBConstraints(2, 2).anchorEast());
            spnIQ = new JNumberSpinner(PKMN_IQ_MAX);
            statsPanel.add(spnIQ, new GBConstraints(3, 2).fillHorizontal());

            statsPanel.add(new JLabel("Attack:"), new GBConstraints(0, 3).anchorEast());
            spnAtk = new JNumberSpinner(PKMN_STAT_MAX);
            statsPanel.add(spnAtk, new GBConstraints(1, 3).fillHorizontal());

            statsPanel.add(new JLabel("Defense:"), new GBConstraints(2, 3).anchorEast());
            spnDef = new JNumberSpinner(PKMN_STAT_MAX);
            statsPanel.add(spnDef, new GBConstraints(3, 3).fillHorizontal());

            statsPanel.add(new JLabel("Sp. Atk:"), new GBConstraints(0, 4).anchorEast());
            spnSpAtk = new JNumberSpinner(PKMN_STAT_MAX);
            statsPanel.add(spnSpAtk, new GBConstraints(1, 4).fillHorizontal());

            statsPanel.add(new JLabel("Sp. Def:"), new GBConstraints(2, 4).anchorEast());
            spnSpDef = new JNumberSpinner(PKMN_STAT_MAX);
            statsPanel.add(spnSpDef, new GBConstraints(3, 4).fillHorizontal());

            statsPanel.add(new JLabel("Evo Lvl 1:"), new GBConstraints(0, 5).anchorEast());
            spnEvo1Lvl = new JNumberSpinner(PKMN_STAT_MAX);
            statsPanel.add(spnEvo1Lvl, new GBConstraints(1, 5).fillHorizontal());

            statsPanel.add(new JLabel("Evo Lvl 2:"), new GBConstraints(2, 5).anchorEast());
            spnEvo2Lvl = new JNumberSpinner(PKMN_STAT_MAX);
            statsPanel.add(spnEvo2Lvl, new GBConstraints(3, 5).fillHorizontal());
        }
        {
            JPanel flagsPanel = new GBPanel();
            contentPane.add(flagsPanel, new GBConstraints(1, 5, 4, 1).fillBoth());
            {
                GBLayout gbl = new GBLayout();
                gbl.columnWidths(70, 0);
                gbl.rowHeights(0, 0, 0);
                gbl.columnWeights(0.0, 1.0);
                gbl.rowWeights(0.0, 0.0, 0.0);
                flagsPanel.setLayout(gbl);
            }
            flagsPanel.add(new JLabel("Unknown 1:"), new GBConstraints(0, 0).anchorEast());
            {
                JPanel unk1Panel = new JPanel();
                FlowLayout flowLayout = (FlowLayout) unk1Panel.getLayout();
                flowLayout.setAlignment(FlowLayout.LEFT);
                flowLayout.setVgap(0);
                flowLayout.setHgap(0);
                flagsPanel.add(unk1Panel, new GBConstraints(1, 0).fillBoth());
                unk1 = new JBooleanLabel();
                unk1Panel.add(unk1);
            }
            flagsPanel.add(new JLabel("IQ Map:"), new GBConstraints(0, 1, 1, 2).anchorEast());
            {
                JPanel iqMapPanelA = new JPanel();
                FlowLayout fl_iqMapPanelA = (FlowLayout) iqMapPanelA.getLayout();
                fl_iqMapPanelA.setAlignment(FlowLayout.LEFT);
                fl_iqMapPanelA.setVgap(0);
                fl_iqMapPanelA.setHgap(0);
                flagsPanel.add(iqMapPanelA, new GBConstraints(1, 1).fillBoth().insetB(0));
                for(int i=0; i<42; i++) {
                    iqMapList[i] = new JBooleanLabel();
                    iqMapList[i].setToolTipText(SkyIQRes.getInstance().getValue(i));
                    iqMapPanelA.add(iqMapList[i]);
                }
            }
            {
                JPanel iqMapPanelB = new JPanel();
                FlowLayout fl_iqMapPanelB = (FlowLayout) iqMapPanelB.getLayout();
                fl_iqMapPanelB.setAlignment(FlowLayout.LEFT);
                fl_iqMapPanelB.setVgap(0);
                fl_iqMapPanelB.setHgap(0);
                flagsPanel.add(iqMapPanelB, new GBConstraints(1, 2).fillBoth());
                for(int i=42; i<iqMapList.length; i++) {
                    iqMapList[i] = new JBooleanLabel();
                    iqMapList[i].setToolTipText(SkyIQRes.getInstance().getValue(i));
                    iqMapPanelB.add(iqMapList[i]);
                }
            }
        }
        {
            contentPane.add(new JLabel("Tactic:"), new GBConstraints(1, 6).anchorEast());
            cmbTactic = new JComboBox<PropertyItem>();
            cmbTactic.setModel(new ArrayListModel<PropertyItem>(SkyTacticRes.getInstance().getPropertyList()));
            contentPane.add(cmbTactic, new GBConstraints(2, 6, 3, 1).fillHorizontal());
        }
        {
            JPanel attacksPanel = new JPanel();
            attacksPanel.setLayout(new GridLayout(4, 1, 5, 5));
            contentPane.add(attacksPanel, new GBConstraints(1, 7, 4, 1).fillBoth());
            {
                attackEditPanels = new AttackEditPanel[PKMN_NUM_ATTACKS];
                for(int i=0; i<PKMN_NUM_ATTACKS; i++) {
                    attackEditPanels[i] = new AttackEditPanel();
                    attacksPanel.add(attackEditPanels[i]);
                }
            }
        }
        {
            JPanel buttonPane = new JPanel();
            FlowLayout flowLayout = (FlowLayout) buttonPane.getLayout();
            flowLayout.setAlignment(FlowLayout.RIGHT);
            contentPane.add(buttonPane, new GBConstraints(1, 9, 4, 1).fillBoth());
            {
                JButton btnImportPkmn = new JButton("Import");
                btnImportPkmn.setActionCommand("Import Pokemon");
                btnImportPkmn.addActionListener(this.actionListener);
                buttonPane.add(btnImportPkmn);
            }
            {
                JButton btnExportPkmn = new JButton("Export");
                btnExportPkmn.setActionCommand("Export Pokemon");
                btnExportPkmn.addActionListener(this.actionListener);
                buttonPane.add(btnExportPkmn);
            }
            {
                JButton btnUpdatePkmn = new JButton(updatePokemonAction);
                buttonPane.add(btnUpdatePkmn);
                frame.getRootPane().setDefaultButton(btnUpdatePkmn);
            }
        }
    }

}
